/* This optimized version handles "nu1". Restricted names are generated using variables.
   nu x P = P.
   P | 0 = 0 | P = P.
*/

/*
	proc_exp --> 	  zero
			| pref(out(X, Y), Exp)
			| pref(in(X, Y), Exp)
			| pref(tau, Exp)
%			| input_choice([..., pref(in(X, Y), Exp), ...])
			| sum(Exp, Exp)
			| choice([..., Exp, ...])
			| prob_choice([..., pref(tau(Prob), Exp), ...])
			| nu(X, Exp)
			| match((X=Y), Exp)
			| par(Exp, Exp)
			| proc(Term)

	proc_def -->	def(Name(Params), Exp)	where Params \subseteq n(Exp)
*/

:- import append/3 from basics.
:- import length/2 from basics.
:- import ith/3 from basics.

%:- table rec_trans/4.

% TAU-PREFIX:
trans(pref(tau, P), [pstep(1, tau, P)], _N, []).

% OUTPUT-PREFIX:
trans(pref(out(X, Y), P), [pstep(1, out(X1, Y), P)], N, V) :-
	((nonvar(X), nonvar(X1)) -> X == X1,
				    V = []
				 ;
				    (N == 1 -> print_match(F, X, X1),
					       X = X1,
					       V = [X, matchf(F)]
					    ;
					       X = X1,
					       V = [])).

% INPUT-PREFIX:
trans(pref(in(X, Y), P), [pstep(1, in(X1, Y), P)], N, V) :-
	((nonvar(X), nonvar(X1)) -> X == X1,
				    V = []
				 ;
				    (N == 1 -> print_match(F, X, X1),
					       X = X1,
					       V = [X, matchf(F)]
					    ;
					       X = X1,
					       V = [])).

% INPUT_CHOICE:
%trans(input_choice(InputBranches), PSteps, N, V) :-
%	length(InputBranches, Size),
%	upto(Size, I),
%	ith(I, InputBranches, InputBranch),
%	trans(InputBranch, PSteps, N, V).

% SUM: BINARY CHOICE:
trans(sum(P, Q), PSteps, N, V) :-
	trans(P, PSteps, N, V).
trans(sum(P, Q), PSteps, N, V) :-
	trans(Q, PSteps, N, V).

% CHOICE:
trans(choice(Branches), PSteps, N, V) :-
	length(Branches, Size),
	upto(Size, I),
	ith(I, Branches, Branch),
	trans(Branch, PSteps, N, V).

%PROB_CHOICE:
trans(prob_choice(ProbBranches), PSteps, _N, V) :-
	prob_branch(ProbBranches, PSteps, V).

prob_branch([], [], []).
prob_branch([pref(tau(FirstProb), P)|Others], PSteps, V) :-
	prob_branch(Others, OtherPSteps, V),
	append([pstep(FirstProb, tau, P)], OtherPSteps, PSteps).

:- import format/2 from format.

/*print_match(N, X, Y) :-
	var(N) -> (format("@~w = ~w~n", [X, Y]), fail).
*/

:- dynamic file_no/1.

:- assert(file_no(0)).

print_match(S, X, Y) :-
	file_no(K),
	K1 is K + 1,
	fmt_write_string(S, ".m%d", args(K1)),
	open(file(S), write, F),
	write(F, X=Y),
	close(F),
	retract(file_no(K)),
	assert(file_no(K1)).
%	tmpfile_open(K),
%	write(K, X=Y).

% MATCH:
trans(match((X=Y), P), PSteps, N, V) :- 
	X == Y -> trans(P, PSteps, N, V)
	       ;
		  (\+(X\=Y) -> (print_match(F, X, Y), % added by wp
			        X = Y,
			        trans(P, PSteps, N, V1),
			        append([X, matchf(F)], V1, V))).

% MISMATCH:
trans(mismatch((X, Y), P), PSteps, N, V) :-
	X \== Y,
	trans(P, PSteps, N, V1),
	append([X, mismatch(Y)], V1, V).

% IDE:
trans(proc(PN), PSteps, N, V) :-
	def(PN, P), 
	P \== proc(PN), 	% peng 15/06/2007
	trans(P, PSteps, N, V).

% PAR:
trans(par(P, Q), PSteps, N, V) :-
	var(PSteps) -> ((trans(P, PPSteps, N, V),
		         set_par_psteps(PPSteps, Q, PSteps, 0));
		        (trans(Q, QPSteps, N, V),
		         set_par_psteps(QPSteps, P, PSteps, 1));
			(comp_trans(par(P, Q), Next, V),
			 PSteps = [pstep(1, tau, Next)]))
		    ;
		       ntrans(par(P, Q), PSteps, N, V).

set_par(P, Q, R) :-
	P == zero -> Q = R
		  ;  (Q == zero -> P = R
				;  R = par(P, Q)).

set_par_psteps([], _, [], _).
set_par_psteps([pstep(Prob, A, P)|Others], Q, PSteps, Which) :-
	set_par_psteps(Others, Q, OtherPSteps, Which),
	(Which == 0 -> set_par(P, Q, R)
		    ;
		       set_par(Q, P, R)),
	append([pstep(Prob, A, R)], OtherPSteps, PSteps).

% COM:
comp_trans(par(P, Q), R, V) :-
	trans(P, [pstep(1, A, P1)], 0, V1),
	complement(A, B),
	trans(Q, [pstep(1, B, Q1)], 1, V2),
	append(V1, V2, V),
	set_par(P1, Q1, R).

complement(in(X, V), out(X, V)).
complement(out(X, V), in(X, V)).

% CLOSE:
comp_trans(par(P, Q), R, V) :-
	trans(P, [pstep(1, A, P1)], 0, V1),
	comp_bound(A, B, W),
	trans(Q, [pstep(1, B, Q1)], 1, V2),
	append(V1, V2, V),
	set_close_par(P1, Q1, W, R).

comp_bound(outbound(X, W), in(X, W), W).
comp_bound(in(X, W), outbound(X, W), W).

comp_trans(par(P, Q), R, V) :-
	trans(P, [pstep(1, A, P1)], 0, V1),
	comp_bound1(A, B, W),
	trans(Q, [pstep(1, B, Q1)], 1, V2),
	append(V1, V2, V),
	set_close_par(P1, Q1, W, R).

comp_bound1(outbound1(X, W), in(X, W), W).
comp_bound1(in(X, W), outbound1(X, W), W).

set_close_par(P, Q, W, R) :-
	functor(W, F, _),
	(F == args) -> W =.. L,
		       close_trans(P, Q, L, R)
		    ;
		       set_par(P, Q, R1),
		       R = nu(W, R1).

close_trans(P, Q, [], R) :-
	set_par(P, Q, R).
close_trans(P, Q, [H|L], R) :-
	compund(H) -> close_trans(P, Q, L, R1),
		      R = nu(H, R1)
		   ;  close_trans(P, Q, L, R).

ntrans(par(P, Q), [pstep(Prob, A, R)], N, V) :-
	trans(P, [pstep(Prob, A, P1)], N, V),
	set_par(P1, Q, R).

ntrans(par(P, Q), [pstep(Prob, A, R)], N, V) :-
	trans(Q, [pstep(Prob, A, Q1)], N, V),
	set_par(P, Q1, R).

% RES:
trans(nu(Y, P), PSteps, N, V) :- 
	gen_new_name(Y),
	var(PSteps) -> trans(P, PPSteps, N, V),
		       not_in_any(Y, PPSteps),
		       notmemberchk(Y, V),
		       set_nu_psteps(PPSteps, Y, PSteps)
		    ;
		       ntrans(nu(Y, P), PSteps, N, V).

set_nu_psteps([], _, []).
set_nu_psteps([pstep(Prob, A, P1)|Others], Y, PSteps) :-
	set_nu_psteps(Others, Y, OtherPSteps),
	append([pstep(Prob, A, nu(Y, P1))], OtherPSteps, PSteps).

trans(nu1(Y, P), PSteps, N, V) :-
	gen_new_name(Y),
	var(PSteps) -> trans(P, PPSteps, N, V), 
		       not_in_any(Y, PPSteps),
		       notmemberchk(Y, V),
		       set_nu1_psteps(PPSteps, Y, PSteps)
		    ;
		       ntrans(nu1(Y, P), PSteps, N, V).

set_nu1(Y, P, Pnew) :-
	contains(P, Y) -> Pnew = nu1(Y, P)
		       ;
			  Pnew = P.

set_nu1_psteps([], _, []).
set_nu1_psteps([pstep(Prob, A, P1)|Others], Y, PSteps) :-
	set_nu1_psteps(Others, Y, OtherPSteps),
	set_nu1(Y, P1, Pnew),
	append([pstep(Prob, A, Pnew)], OtherPSteps, PSteps).

not_in_any(_, []).
not_in_any(Z, [pstep(_, A, _)|L]) :-
	not_in(Z, A),
	not_in_any(Z, L).

not_in(Z, in(X,Y)) :- Z\==X, varnchk(Z, Y). %Z\==Y.
not_in(Z, out(X,Y)) :- Z\==X, varnchk(Z, Y). %Z\==Y.
not_in(Z, outbound(X,Y)) :- Z\==X, varnchk(Z, Y). %Z\==Y.
not_in(Z, outbound1(X,Y)) :- Z\==X, varnchk(Z, Y). %Z\==Y.
not_in(_, tau).

notmemberchk(_X, []).
%notmemberchk(X, [Y|L]) :- X\==Y, notmemberchk(X, L). % Nov. 14, 2006
notmemberchk(X, [Y|L]) :-
	varnchk(X, Y),
	notmemberchk(X, L).

varnchk(X, Y) :-
	(compound(Y), functor(Y, F, _), F == args) -> Y =.. L, varnchk_list(X, L)
						   ;  X \== Y.

varnchk_list(_, []).
varnchk_list(X, [H|L]) :-
	X \== H,
	varnchk_list(X, L).

varchk(Y, Z) :-
	(compound(Z), functor(Z, F, N), F == args) -> upto(N, I), arg(I, Z, Z1), Y == Z1
		   				   ;  Y == Z.

ntrans(nu(Y, P), [pstep(Prob, A, nu(Y, P1))], N, V) :-
	trans(P, [pstep(Prob, A, P1)], N, V),
	not_in(Y, A),
	notmemberchk(Y, V).

ntrans(nu1(Y, P), [pstep(Prob, A, Pnew)], N, V) :-
	trans(P, [pstep(Prob, A, P1)], N, V),
	not_in(Y, A),
	notmemberchk(Y, V),
	set_nu1(Y, P1, Pnew).

% OPEN:
trans(nu(Y, P), PSteps, N, V) :-
	res_trans(nu(Y, P), PSteps, N, V).

res_trans(nu(Y, P), [pstep(1, outbound(X, Z), P1)], N, V) :-
	gen_new_name(Y),
	(trans(P, [pstep(1, out(X, Z), P1)], N, V);
	 trans(P, [pstep(1, outbound(X, Z), P1)], N, V)),
%	Y == Z,	% if Z is a variable, then it is an implicit parameter
%		% so cannot match a channel name!!
	varchk(Y, Z),
	Y \== X,
	notmemberchk(Y, V).

trans(nu1(Y, P), PSteps, N, V) :-
	res_trans(nu1(Y, P), PSteps, N, V).

res_trans(nu1(Y, P), [pstep(1, outbound1(X, Z), P1)], N, V) :-
	gen_new_name(Y),
	(trans(P, [pstep(1, out(X, Z), P1)], N, V);
	 trans(P, [pstep(1, outbound1(X, Z), P1)], N, V)),
%	Y == Z,	% if Z is a variable, then it is an implicit parameter
%		% so cannot match a channel name!!
	varchk(Y, Z),
	Y \== X,
	notmemberchk(Y, V).

gen_new_name(N):-
	N = name(_).

contains(T, X) :-
        (T = name(_) -> T == X
		     ; 
			functor(T,_,N), 
			upto(N,I), 
			arg(I,T,T1), 
			contains(T1,X)
        ).
 
upto(N, N) :- N > 0.
upto(N, I) :- N > 0, N1 is N-1, upto(N1, I).
