#!/bin/sh
#
# File: PROBPI/install.sh
#
# Script to install the MMC_SP system.
# 

# You may want to change the default setting of the following three
# variables before you run this script (so you do not have to type in
# them interactively).

# ===============================================
# You SHOULD NOT change anything below this line.
# ===============================================

MMC_SP_VERSION=ProNoBis
MMC_SP_URL=http://www.cs.ucl.ac.uk/staff/p.wu/index.html#mmc_sp
MMC_SP_PATH=`pwd`
INSTALL_LOG="`pwd`/Installation_summary"

# Some utility functions
# --------------------------------------------
get_str()
{
    $ECHO "$* \c"
    read ANS
}

# Set ECHO
# --------------------------------------------
# Use Sysv echo if possible
if [ -x /usr/5bin/echo ]
then
    ECHO=/usr/5bin/echo				# SunOS SysV echo
elif [ -z "`(echo -e a) 2>/dev/null | fgrep e`" ]
then
    ECHO="echo -e"				# GNU bash, etc.
else
    ECHO=echo					# generic SysV
fi
if [ "$OS_SYS" = QNX ]
then
    ECHO=echo
fi

cat <<EOF
========================================================================
 Before running the installation, you must have installed
       XSB -- a tabled logic programming system, and
 Please see the README file for details on obtaining and installing the 
 packages.
========================================================================
EOF

while :
do
    get_str "Continue? (y/n)\n  [y]"
    if [ -z "$ANS" -o "$ANS" = "y" -o "$ANS" = "n" -o \
	 "$ANS" = "Y" -o "$ANS" = "N" ]; then
	break
    fi
done

$ECHO

if [ "$ANS" = "n" -o "$ANS" = "N" ]; then
    exit 0
fi

$ECHO "\nInstalling MMC_SP version $MMC_SP_VERSION"
$ECHO "======================================="
$ECHO "\nPlease visit the MMC_SP web site: $MMC_SP_URL"

# Get the full path name of the XSB executable
# --------------------------------------------
out=`which xsb 2>&1`
case "$out" in
    *xsb)	XSB_EXEC="$out";;
    which*)	XSB_EXEC="/usr/bin/xsb";;
esac
while :
do
    get_str "\nEnter the full path name of the XSB executable\n  [${XSB_EXEC}]:"
    if [ -z "$ANS" ]
    then
	ANS=$XSB_EXEC
    fi

    if [ -x "$ANS" ]
    then
	XSB_EXEC=$ANS
	break
    else
	$ECHO "  Error: $ANS does not exist, please try again."
    fi
done


rm -f $INSTALL_LOG

$ECHO "\nMMC_SP Installation Summary (`date`)" > $INSTALL_LOG
$ECHO "-------------------------------------------------------" >> $INSTALL_LOG
$ECHO "`uname -a`\n" >> $INSTALL_LOG
$ECHO "XSB_EXEC=$XSB_EXEC" >> $INSTALL_LOG
$ECHO "MMC_SP_PATH=$MMC_SP_PATH" >> $INSTALL_LOG


# Create startup script
# --------------------------------------------
cd System
$ECHO "\nCreating Prolog components..." | tee -a $INSTALL_LOG
rm -f probmmc.P
$ECHO ":- assert(mmc_sp_directory('$MMC_SP_PATH/System'))." > probmmc.P
cat probmmc.proto >> probmmc.P

$ECHO "Creating startup script..." | tee -a $INSTALL_LOG

cd ..
if [ ! -d bin ] ; then
    mkdir bin
fi
cd bin
rm -f probmmc

$ECHO "#!/bin/sh" > probmmc
$ECHO "## This file is automatically generated by install.sh" >> probmmc
$ECHO >> probmmc
$ECHO "$XSB_EXEC -e \"consult('$MMC_SP_PATH/System/probmmc').\" \$*" >> probmmc

$ECHO "#!/bin/csh" > pi2mmc
$ECHO "(\\" >> pi2mmc
$ECHO "echo \"consult('\$1').\";\\" >> pi2mmc
$ECHO "echo \"stg(\$2).\"\\" >> pi2mmc
$ECHO ")\\" >> pi2mmc
$ECHO "|$MMC_SP_PATH/bin/probmmc|sed \"s/| ?- //g\"|sed \"s/^yes//\"" >> pi2mmc

chmod +x probmmc
chmod +x pi2mmc

# Make Prolog components
# --------------------------------------------
cd ../System
$ECHO "Creating all other necessary components..." | tee -a $INSTALL_LOG

XSB=$XSB_EXEC 2>&1 | tee -a $INSTALL_LOG

$ECHO "Installation finished." >> $INSTALL_LOG
$ECHO "-------------------------------------------------------" >> $INSTALL_LOG

$ECHO
$ECHO "Installation done.  Now you can run MMC_SP using the shell script:"
$ECHO "\t${MMC_SP_PATH}/bin/probmmc"
